
// FILE : link.h											build 3
// =================================================================
// Linked-List Template : LinkedList
// =================================================================
// Copyright (c) 1999, 2000 by jhkim (neuron)
//
//
// LinkedList manages a linked list which contains formal data 
// of 8, 16, 32bits in size.
// 
// CONTACT TO
// e-mail	: cherishee@hotmail.com
// office	: +82-361-240-1457 (Graphics and Virtual Environment Lab. in Hallym univ.)
// cellular	: 011-9945-2575
//
// NOTE THAT
// All the following codes are contributed with NO-WARRANTY!!!
// Please use and modify it freely at your own risk.
//
// - fixed: 
// this source is safe in in-process multi-threaded 
// environment. not support simultaneous access from 
// multi-processes.
//														2000/2/2
// - added:
// you can now retrieve the total imported number of data
// with GetNumOfData().
//														2000/2/2
// - added:
// Delete(LINKEDLIST *item)
//														2000/5/22
// - fixed:
// a small bug in Delete() is fixed.
//														2000/5/22


#ifndef _LINKEDLIST_H
#define _LINKEDLIST_H

#pragma message (" + " __FILE__)

#include <windows.h>

#define DECL_TEMPLATE	template <class datatype>
#define DECL_CLASS		LinkedList<datatype>
#define DECL_PROTOC		DECL_CLASS DECL_CLASS	// returns class
#define DECL_PROTO		DECL_CLASS


DECL_TEMPLATE
class LinkedList
{

public:

	typedef struct LINKEDLISTTAG
	{
		datatype v;
		struct LINKEDLISTTAG *next;
	} LINKEDLIST;


	LinkedList();
	~LinkedList();

	LINKEDLIST* Add(datatype data);

	LINKEDLIST* Find(datatype data);
	LINKEDLIST* FindPrev();
	LINKEDLIST* FindPrev(datatype data);
	LINKEDLIST* FindFirst();
	LINKEDLIST* FindNext(LINKEDLIST *pLL);

	BOOL Delete(LINKEDLIST *item);
	BOOL Delete(datatype data);
	BOOL DeleteFirst();
	BOOL DeleteAll();

	BOOL Change(datatype dest, datatype data);
	BOOL Insert(datatype dest, datatype data);

	datatype GetData(LINKEDLIST *pLL);
	WORD GetNumOfData();

private:

	CRITICAL_SECTION	_mlink_cs0;
	LINKEDLIST *pFirst, *pLast, *pPrev;
	WORD	data_counter;
};

DECL_TEMPLATE
DECL_PROTO::LinkedList()
{
	InitializeCriticalSection(&_mlink_cs0);

	pFirst = NULL;
	pLast = NULL;
	pPrev = NULL;

	data_counter = 0;
}

DECL_TEMPLATE
DECL_PROTO::~LinkedList()
{
	DeleteAll();

	DeleteCriticalSection(&_mlink_cs0);
}

DECL_TEMPLATE
DECL_PROTO::LINKEDLIST* DECL_CLASS::Add(datatype data)
{
	LINKEDLIST *item;

	item = (LINKEDLIST*)malloc(sizeof(LINKEDLIST));

	EnterCriticalSection(&_mlink_cs0);

	item->v = data;
	item->next = NULL;

	if (pFirst == NULL)
	{
		pFirst = item;
		pLast = item;
	}
	else
	{
		pLast->next = item;
		pLast = item;
	}

	data_counter++;

	LeaveCriticalSection(&_mlink_cs0);
	return item;
}

DECL_TEMPLATE
DECL_PROTO::LINKEDLIST* DECL_CLASS::Find(datatype data)
{
	LINKEDLIST *item;

	EnterCriticalSection(&_mlink_cs0);

	pPrev = NULL;
	item = pFirst;

	while (item != NULL)
	{
		if (item->v == data)
		{
			LeaveCriticalSection(&_mlink_cs0);
			return item;
		}

		pPrev = item;	// save the current item

		item = item->next;
	}

	LeaveCriticalSection(&_mlink_cs0);
	return NULL;
}

DECL_TEMPLATE
DECL_PROTO::LINKEDLIST* DECL_CLASS::FindPrev()
{
	return pPrev;
}

DECL_TEMPLATE
DECL_PROTO::LINKEDLIST* DECL_CLASS::FindPrev(datatype data)
{
	Find(data);
	return pPrev;
}

DECL_TEMPLATE
DECL_PROTO::LINKEDLIST* DECL_CLASS::FindFirst()
{
	return pFirst;
}

DECL_TEMPLATE
DECL_PROTO::LINKEDLIST* DECL_CLASS::FindNext(LINKEDLIST *pLL)
{
	EnterCriticalSection(&_mlink_cs0);

	if (pLL == NULL)
	{
		LeaveCriticalSection(&_mlink_cs0);
		return NULL;
	}

	pPrev = pLL;

	LeaveCriticalSection(&_mlink_cs0);
	return pLL->next;
}

DECL_TEMPLATE
BOOL DECL_CLASS::Delete(LINKEDLIST *item)
{
	LINKEDLIST *prev;

	EnterCriticalSection(&_mlink_cs0);
	prev = FindPrev();

	if (item == NULL)
	{	// no such a data
		LeaveCriticalSection(&_mlink_cs0);
		return FALSE;
	}

	if (item == pFirst)
	{
		pFirst = FindNext(pFirst);
		free(item);
//		SafeDelete<LINKEDLIST>(&item);

		data_counter--;
		LeaveCriticalSection(&_mlink_cs0);
		return TRUE;
	}

	if (item == pLast)
	{
		prev->next = NULL;
		pLast = prev;
		free(item);
//		SafeDelete<LINKEDLIST>(&item);

		data_counter--;
		LeaveCriticalSection(&_mlink_cs0);
		return TRUE;
	}

	prev->next = item->next;
	free(item);
//	SafeDelete<LINKEDLIST>(&item);

	data_counter--;
	LeaveCriticalSection(&_mlink_cs0);
	return TRUE;
}

DECL_TEMPLATE
BOOL DECL_CLASS::Delete(datatype data)
{
	LINKEDLIST *item, *prev;

	item = Find(data);

	return Delete(item);
}

DECL_TEMPLATE
BOOL DECL_CLASS::DeleteFirst()
{
	LINKEDLIST *item;

	EnterCriticalSection(&_mlink_cs0);

	item = FindFirst();

	if (item)
	{
		// move pFirst(list index) to the next linked position.
		pFirst = FindNext(pFirst);
		free(item);
//		SafeDelete<LINKEDLIST>(&item);

		data_counter--;
		LeaveCriticalSection(&_mlink_cs0);
		return TRUE;
	}

	// the requested data was not found.
	LeaveCriticalSection(&_mlink_cs0);
	return FALSE; // no data to delete
}

DECL_TEMPLATE
BOOL DECL_CLASS::DeleteAll()
{
	LINKEDLIST *item;

	EnterCriticalSection(&_mlink_cs0);

	item = FindFirst();

	if (item == NULL)
	{
		LeaveCriticalSection(&_mlink_cs0);
		return FALSE;
	}

	while ((item = FindNext(item)))
		free(FindPrev());
//		delete FindPrev();

	free(FindPrev());
//	delete FindPrev();

	pFirst = NULL;
	pLast = NULL;

	data_counter = 0;
	LeaveCriticalSection(&_mlink_cs0);
	return TRUE;
}

DECL_TEMPLATE
BOOL DECL_CLASS::Change(datatype dest, datatype data)
{
	LINKEDLIST *item;

	// find the destinative item; it also guarantees mutual-exclusion scheme.
	item = Find(dest);

	EnterCriticalSection(&_mlink_cs0);

	if (item == NULL)
	{
		LeaveCriticalSection(&_mlink_cs0);
		return FALSE;
	}

	if (item)
		item->v = data;

	LeaveCriticalSection(&_mlink_cs0);
	return TRUE;
}

DECL_TEMPLATE
BOOL DECL_CLASS::Insert(datatype dest, datatype data)
{
	LINKEDLIST *item, *last;
	LINKEDLIST *tmp;

	// find the destinative item; it also guarantees mutual-exclusion scheme.
	item = Find(dest);

	EnterCriticalSection(&_mlink_cs0);

	if (item == NULL)
	{
		LeaveCriticalSection(&_mlink_cs0);
		return FALSE; // no destination
	}

	// make a new data
	tmp = (LINKEDLIST*)malloc(sizeof(LINKEDLIST));
	tmp->v = data;

	if (item == FindFirst())
	{
		tmp->next = item;
		pFirst = tmp;
		LeaveCriticalSection(&_mlink_cs0);
		return TRUE;
	}

	tmp->next = FindPrev()->next;
	FindPrev()->next = tmp;

	LeaveCriticalSection(&_mlink_cs0);
	return TRUE;
}

DECL_TEMPLATE
datatype DECL_CLASS::GetData(LINKEDLIST *pLL)
{
	if (pLL == NULL)
		return (datatype)0;

	return pLL->v;
}

DECL_TEMPLATE
WORD DECL_CLASS::GetNumOfData()
{
	return data_counter;
}

#endif

